"""
Author: Pierre Tocquin (University of Liège, ptocquin@uliege.be)
Description: This script automates the processing of CZI images by converting them to TIFF,
             analyzing them with Ilastik, and resizing the input and segmented outputs.
             
             The script uses a YAML configuration file to define global parameters such as:
             - Output directory for processed files (`output_dir`)
             - Compression type for TIFF files (`compression`)
             - Resize factor for scaling images (`resize_factor`)
             - Path to the Ilastik project file (`project`)
             - Path to the Ilastik executable (`ilastik`)
             - Minimum size of red clusters to preserve as true nodes of Ranvier (`area_threshold`)
             - Kernel size for filling interruptions in nerve segments (`closing_kernel_size`)
             - Minimum and maximum aspect ratios to identify true nodes of Ranvier (`aspect_ratio_min`, `aspect_ratio_max`)

Usage: python process_pipeline.py file1.czi file2.czi ... --config /path/to/config.yaml

Example configuration file (config.yaml):
    output_dir: ./output
    compression: lzw
    resize_factor: 0.25
    project: /path/to/ilastik_project.ilp
    ilastik: /path/to/ilastik_executable
    area_threshold: 50
    closing_kernel_size: 11
    aspect_ratio_min: 1.5
    aspect_ratio_max: 5

Features:
    - Assembles mosaics from CZI tiles into single TIFF files.
    - Performs image segmentation using Ilastik in headless mode.
    - Corrects segmentation issues in the segmented image:
        * Preserves true nodes of Ranvier by filtering red clusters based on size and aspect ratio.
        * Applies morphological operations to remove noise (e.g., thin structures).
        * Fills interruptions in nerve segments caused by white gaps.
    - Resizes both original and segmented images for downstream analysis.
    - Removes intermediate files to optimize disk usage.

Dependencies:
    - PyYAML: For reading the YAML configuration file.
      Installation: pip install pyyaml
    - aicspylibczi: For processing CZI files.
      Installation: pip install aicspylibczi
    - tifffile: For handling TIFF files.
      Installation: pip install tifffile
    - OpenCV: For image resizing and morphological operations.
      Installation: pip install opencv-python-headless
    - Ilastik: For segmentation in headless mode (requires Ilastik installation).

License: MIT License
Copyright (c) 2024 Pierre Tocquin
"""

import argparse
import os
import cv2
import subprocess
import yaml
import tifffile as tiff
import numpy as np
from aicspylibczi import CziFile


def load_config(config_path):
    """
    Loads and validates the configuration from a YAML file.

    Args:
        config_path (str): Path to the YAML configuration file.

    Returns:
        dict: Dictionary of configuration options.
    """
    with open(config_path, 'r') as file:
        config = yaml.safe_load(file)
    
    required_keys = [
        'output_dir', 'compression', 'resize_factor', 
        'project', 'ilastik', 'area_threshold', 
        'closing_kernel_size', 'aspect_ratio_min', 'aspect_ratio_max'
    ]    
    for key in required_keys:
        if key not in config:
            raise ValueError(f"Missing required configuration key: {key}")
    
    return config


def process_czi_file(file_path, config):
    """
    Processes a CZI file, assembles mosaics, analyzes them with Ilastik, and corrects segmentation issues.
    Each series within the CZI file is processed sequentially.

    Workflow:
        1. Assemble mosaic from CZI tiles into a single TIFF file.
        2. Process the mosaic with Ilastik for segmentation.
        3. Correct segmentation issues in the segmented image:
            - Preserve true nodes of Ranvier by filtering red clusters based on size and aspect ratio.
            - Apply morphological operations to smooth and refine segmentation results.
        4. Resize both the original mosaic and segmented outputs.
        5. Remove intermediate files to optimize disk usage.

    Args:
        file_path (str): Path to the CZI file to process.
        config (dict): Configuration options loaded from YAML, including:
            - output_dir (str): Directory for saving processed files.
            - compression (str): Compression type for TIFF files (e.g., "lzw").
            - resize_factor (float): Scaling factor for resizing images.
            - project (str): Path to the Ilastik project file.
            - ilastik (str): Path to the Ilastik executable.
            - area_threshold (int): Minimum size for preserving red clusters as nodes of Ranvier.
            - closing_kernel_size (int): Kernel size for filling interruptions in blue regions.
            - aspect_ratio_min (float): Minimum aspect ratio for identifying nodes of Ranvier.
            - aspect_ratio_max (float): Maximum aspect ratio for identifying nodes of Ranvier.
    """

    print(f"Processing file: {file_path}")

    # Load the CZI file
    czi = CziFile(file_path)
    dims = czi.dims  # Example: "HSCMYXA"
    x_idx = dims.index('X')
    y_idx = dims.index('Y')

    # Extract the file name without the extension
    input_filename = os.path.splitext(os.path.basename(file_path))[0]

    num_series = len(czi.get_dims_shape())
    print(f"Number of detected series (S): {num_series}")

    for s in range(num_series):
        print(f"  Processing series {s}")

        # Retrieve information about the tiles for this series
        tiles_nb = len(czi.get_all_mosaic_tile_bounding_boxes(S=s, C=0, A=0, H=0))
        print(f"    Number of tiles for S={s}: {tiles_nb}")

        # Initialize the global dimensions of the mosaic
        max_x, max_y = 0, 0
        tiles = []

        # Iterate through the tiles and retrieve their data
        for m in range(tiles_nb):
            # Read the tile
            tile, _ = czi.read_image(S=s, C=0, M=m, A=0, H=0)

            # Retrieve the tile's coordinates and dimensions
            tile_boundings = czi.get_mosaic_tile_bounding_box(S=s, C=0, M=m, A=0, H=0)

            # Use the first tile's coordinates to adjust the offsets
            if m == 0:
                shift_x = tile_boundings.x
                shift_y = tile_boundings.y

            # Compute the corrected positions
            pos_x = tile_boundings.x - shift_x
            pos_y = tile_boundings.y - shift_y
            width, height = tile_boundings.w, tile_boundings.h

            # Add the tile data to the list
            tiles.append((tile, pos_x, pos_y, width, height))

            # Update the global dimensions
            max_x = max(max_x, pos_x + width)
            max_y = max(max_y, pos_y + height)

        # Create an empty image for the final mosaic
        mosaic_image = np.zeros((max_y, max_x), dtype=tiles[0][0].dtype)

        # Assemble the tiles into the final image
        for tile, pos_x, pos_y, width, height in tiles:
            slices = [slice(None)] * len(dims)
            slices[x_idx] = slice(0, width)
            slices[y_idx] = slice(0, height)
            mosaic_image[pos_y:pos_y + height, pos_x:pos_x + width] = tile[0, 0, 0, 0, :, :, 1]

        # Save the assembled mosaic as a TIFF with a series suffix
        output_filename = os.path.join(config['output_dir'], f"{input_filename}_series_{s}.tif")
        tiff.imwrite(output_filename, mosaic_image, compression=config['compression'])

        print(f"Generated TIFF for series {s}: {output_filename}")

        # Process the generated TIFF with Ilastik
        segmented_path = run_ilastik(output_filename, config['project'], config['ilastik'])

        # Load segmented image for correction
        segmented_image = tiff.imread(segmented_path)

        # Correct segmentation problems (red clusters and interruptions)
        corrected_image = correct_segmentation_issues(
            segmented_image,
            area_threshold=config['area_threshold'],
            closing_kernel_size=config['closing_kernel_size'],
            aspect_ratio_min=config['aspect_ratio_min'],
            aspect_ratio_max=config['aspect_ratio_max']
        )

        # Save corrected image
        tiff.imwrite(segmented_path, corrected_image.astype(np.uint8))
        print(f"Corrected segmentation issues in segmented image: {segmented_path}")


        # Resize the original TIFF and the segmented output
        resize_image(output_filename, config['resize_factor'], cv2.INTER_AREA)
        resize_image(segmented_path, config['resize_factor'], cv2.INTER_NEAREST_EXACT)

        # Remove the intermediate TIFF
        os.remove(output_filename)
        print(f"Deleted intermediate TIFF: {output_filename}")


def run_ilastik(input_path, project_path, ilastik_path, output_format="tif", export_source="Simple Segmentation"):
    """
    Runs Ilastik in headless mode on the given input file.
    """
    output_filename_format = os.path.join(
        os.path.dirname(input_path), f"{os.path.splitext(os.path.basename(input_path))[0]}_segmented"
    )
    cmd = [
        ilastik_path,
        "--headless",
        f"--project={project_path}",
        f"--output_format={output_format}",
        f"--export_source={export_source}",
        f"--output_filename_format={output_filename_format}",
        f"--raw-data={input_path}"
    ]
    try:
        subprocess.run(cmd, check=True)
        segmented_path = f"{output_filename_format}.{output_format}"
        print(f"Processed: {input_path} -> {segmented_path}")
        return segmented_path
    except subprocess.CalledProcessError as e:
        print(f"Error processing: {input_path}")
        raise e


def resize_image(image_path, scale_factor, interpolation):
    """
    Resizes an image using OpenCV.
    """
    image = tiff.imread(image_path)
    if image is None:
        raise FileNotFoundError(f"Could not read image: {image_path}")
    
    new_size = (int(image.shape[1] * scale_factor), int(image.shape[0] * scale_factor))
    resized_image = cv2.resize(image, new_size, interpolation=interpolation)

    # Save resized image
    file_dir, file_name = os.path.split(image_path)
    file_base, file_ext = os.path.splitext(file_name)
    resized_path = os.path.join(file_dir, f"{file_base}_resized{file_ext}")
    cv2.imwrite(resized_path, resized_image)
    print(f"Saved resized image: {resized_path}")
    return resized_path


def correct_segmentation_issues(image, area_threshold=50, closing_kernel_size=11, aspect_ratio_min=1.5, aspect_ratio_max=5):
    """
    Corrects segmentation issues in the processed image:
    1. Preserves large red clusters (true nodes of Ranvier) based on area and shape.
    2. Fills interruptions in nerve segments by closing small gaps in blue regions.

    Args:
        image (numpy.ndarray): The segmented image where white (1), blue (2), and red (3) represent classes.
        area_threshold (int): Minimum area to preserve a red cluster as a true node of Ranvier.
        closing_kernel_size (int): Size of the morphological closing kernel (must be odd).
        aspect_ratio_min (float): Minimum aspect ratio to preserve a cluster as a true node of Ranvier.
        aspect_ratio_max (float): Maximum aspect ratio to preserve a cluster as a true node of Ranvier.

    Returns:
        numpy.ndarray: Corrected segmented image.
    """
    # Step 1: Preserve large red clusters (true nodes of Ranvier)
    red_mask = (image == 3).astype(np.uint8)
    num_labels, labels, stats, _ = cv2.connectedComponentsWithStats(red_mask, connectivity=8)

    true_red_clusters = set()
    for label in range(1, num_labels):
        area = stats[label, cv2.CC_STAT_AREA]
        if area < area_threshold:
            continue  # Ignore small clusters

        # Extract the mask for the current cluster
        cluster_mask = (labels == label).astype(np.uint8)

        # Calculate bounding box and aspect ratio
        contours, _ = cv2.findContours(cluster_mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        if not contours:
            continue

        # Use the largest contour for the cluster
        contour = max(contours, key=cv2.contourArea)

        # Calculate the oriented bounding box
        rect = cv2.minAreaRect(contour)
        width, height = rect[1]  # Dimensions of the bounding box

        if width == 0 or height == 0:
            continue  # Skip invalid clusters

        # Calculate aspect ratio
        aspect_ratio = max(width, height) / min(width, height)
        if aspect_ratio_min <= aspect_ratio <= aspect_ratio_max:
            true_red_clusters.add(label)

    # Iterate over all red pixels to remove small red clusters
    red_pixels = np.argwhere(image == 3)
    for y, x in red_pixels:
        if labels[y, x] not in true_red_clusters:
            # Convert small red clusters to blue or white based on neighborhood
            y1, y2 = max(0, y - 1), min(image.shape[0], y + 2)
            x1, x2 = max(0, x - 1), min(image.shape[1], x + 2)
            neighborhood = image[y1:y2, x1:x2]
            blue_count = np.sum(neighborhood == 2)
            white_count = np.sum(neighborhood == 1)

            if blue_count > white_count:
                image[y, x] = 2  # Convert to blue
            else:
                image[y, x] = 1  # Convert to white

    # Step 2: Fill interruptions in blue regions (nerve segments)
    blue_mask = (image == 2).astype(np.uint8)

    # Apply morphological closing to fill small gaps
    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (closing_kernel_size, closing_kernel_size))
    closed_blue_mask = cv2.morphologyEx(blue_mask, cv2.MORPH_CLOSE, kernel)

    # Identify newly filled areas (previously white and now part of the blue region)
    newly_filled = (closed_blue_mask == 1) & (image == 1)
    image[newly_filled] = 2  # Convert these pixels to blue

    # Apply a median filter to smooth the final image
    image = cv2.medianBlur(image, 7)

    return image


def main():
    parser = argparse.ArgumentParser(description="Automate CZI to Ilastik processing pipeline.")
    parser.add_argument("files", nargs='+', help="List of CZI files to process.")
    parser.add_argument("--config", required=True, help="Path to the YAML configuration file.")
    args = parser.parse_args()

    # Load configuration
    config = load_config(args.config)

    # Ensure output directory exists
    if not os.path.isdir(config['output_dir']):
        os.makedirs(config['output_dir'])

    # Process each file through the pipeline
    for file_path in args.files:
        if not os.path.isfile(file_path):
            print(f"File not found: {file_path}")
            continue
        process_czi_file(file_path, config)


if __name__ == "__main__":
    main()
