#!/bin/bash

################################################################################
# Script Name: batch_ilastik.sh
# Description: Automates batch processing of image files using Ilastik's 
#              headless mode. Applies pixel classification model to multiple
#              input files and exports the results in a specified format.
# Author: Pierre TOCQUIN, University of Liege (Belgium), ptocquin@uliege.be
# Date: 2024-12-01
# Requirements: Ilastik version 1.4 or later.
#
# Usage:
#   ./batch_ilastik.sh file1 file2 ... fileN
# Example:
#   ./batch_ilastik.sh ../images/img1.tif ../images/img2.tif
#
# Arguments:
#   file1, file2, ... fileN - List of image files to process.
#
# Output:
#   - Results will be saved in the same directory as the input files, with
#     "_seg" appended to the original filename.
#
# Note:
#   Ensure the Ilastik project file (.ilp) is correctly configured and accessible.
#
# License:
# ------------------------------------------------------------------------------
# MIT License
#
# Copyright (c) 2024 Pierre Tocquin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ------------------------------------------------------------------------------
################################################################################

# Check if at least one argument (file) is provided
if [ "$#" -lt 1 ]; then
  echo "Usage: $0 file1 file2 ... fileN"
  exit 1
fi

# Path to Ilastik executable
ILASTIK_PATH="/Applications/ilastik-1.4.0-OSX.app/Contents/ilastik-release/run_ilastik.sh"

# Path to the Ilastik project file
PROJECT_PATH="teasing.ilp"

# Output configuration
OUTPUT_FORMAT="tif"                 # Format of the output files
EXPORT_SOURCE="Simple Segmentation" # Source type for export

# Loop through all provided files
for RAW_FILE in "$@"; do
  # Check if the file exists
  if [ ! -f "$RAW_FILE" ]; then
    echo "File not found: $RAW_FILE"
    continue
  fi

  # Construct the output file path
  OUTPUT_FILENAME_FORMAT='{dataset_dir}/{nickname}_segmented'

  # Execute Ilastik in headless mode
  $ILASTIK_PATH --headless \
    --project="$PROJECT_PATH" \
    --output_format="$OUTPUT_FORMAT" \
    --export_source="$EXPORT_SOURCE" \
    --output_filename_format="$OUTPUT_FILENAME_FORMAT" \
    --raw-data="$RAW_FILE"

  # Check the status of the command
  if [ $? -eq 0 ]; then
    echo "Processed: $RAW_FILE -> ${OUTPUT_FILENAME_FORMAT}.${OUTPUT_FORMAT}"
  else
    echo "Error processing: $RAW_FILE"
  fi
done

echo "Batch processing completed."